
    #	Make an orrery rendering of the solar system from
    #	an extract file.  This is usually run from an extract
    #	with a single integration step, such as with the demo:
    #
    #	    perl orrery.pl 86859_bary_extract.csv
    #
    #	It leaves the .pov file and the rendering in the
    #	current directory so you can twiddle with them.  This
    #	is a hand-crafted hack--you will have to fix the horrid
    #	adjustment I do to line up the orbits with the view plane
    #	or, ideally, do it correctly.  I got tired of fighting
    #	POV-Ray coordinate systems for a one-off picture.
    
    #	by John Walker
    #	http://www.fourmilab.ch/
    
    #	For details, please see: http://www.fourmilab.ch/documents/canon_transits/

    use strict;
    use warnings;
    
    use Math::VectorReal;
    use Math::Trig;
    
    #	Inflate the sizes of planets by this factor to render them more visible
    my $inflate_planets = 2500;
    
    #	Specify animation frame size
    my ($fwidth, $fheight) = (1024, 768);
    
    my $pvopts = "-v -w$fwidth -h$fheight -d -q9";  # POV-Ray rendering options
    
    my $au = 149597870691.0;	    	# Astronomical unit in metres
        
    my @planets = (
    	"Sun",	    	    #  0
    	"Mercury",  	    #  1
	"Venus",    	    #  2
    	"Earth",    	    #  3
	"Moon",     	    #  4
	"Mars",     	    #  5
	"Jupiter",  	    #  6
	"Saturn",   	    #  7
	"Uranus",   	    #  8
     	"Neptune",  	    #  9
	"Pluto" );  	    # 10

    my %planet_col = (
    	"Sun", "White",
    	"Mercury", "Brown",
	"Venus", "Gray75",
    	"Earth", "Blue",
	"Moon", "Gray75",
	"Mars", "Red",
	"Jupiter", "Orange",
	"Saturn", "Yellow",
	"Uranus", "Cyan",
     	"Neptune", "Blue",
	"Pluto", "Yellow" );
	
    my %planet_radius = (
    	"Sun", 6.96e8 / $au,
    	"Mercury", 2439700 / $au,
	"Venus", 6051800 / $au,
    	"Earth", 4.26352325064817808471E-5,
	"Moon", 1.16178124180819699985E-5,
	"Mars", 3397000 / $au,
	"Jupiter", 71492000 / $au,
	"Saturn", 60268000 / $au,
	"Uranus", 25559000 / $au,
     	"Neptune", 24764000 / $au,
	"Pluto", 1195000 / $au );
	
    my $l;
    
    my ($jd, $yy, $mo, $dd, $hh, $mm, $ss);
    my %planet_vec;
    
    while ($l = <>) {
	#   Load a state vector snapshot into memory
	
	$l =~ m/JD,([^,]+),([^,]+),([^,]+),([^,]+),([^,]+),([^,]+),([^,]+)\s/;
    	($jd, $yy, $mo, $dd, $hh, $mm, $ss) = ($1, $2, $3, $4, $5, $6, $7);
	
	while (1) {
	    $l = <>;
	    if ($l =~ m/^END/) {
	    	last;
	    }
	    $l =~ m/^(\w+),([\d-\.e]+),([\d-\.e]+),([\d-\.e]+)/;
	    $planet_vec{$1} = vector($2, $3, $4);
	}
    	my $fname = sprintf("orrery_%.2f.pov", $jd);
	genscene($fname);
	renderscene($fname);
    }
    	
    sub genscene {
    	my ($fname, $vfi) = @_;
	
	my $sun_pos =  outpos($planet_vec{'Sun'});
	my $jupiter_pos = outpos($planet_vec{'Jupiter'});
	my $vangle = 3;
#$vangle = 0.25;    	# To view inner system
	
    	my $ecliptic_pole = (($planet_vec{'Uranus'} - $planet_vec{'Sun'}) x ($planet_vec{'Saturn'} - $planet_vec{'Sun'}))->norm();
	$ecliptic_pole = $ecliptic_pole * 2000;
	my $fvup = "< " . $ecliptic_pole->x() . ", " . $ecliptic_pole->y() . ", " . $ecliptic_pole->z() . " >";
	
	open(OF, ">$fname");
	
	print OF << "EOD";

#include "colors.inc"

    camera {
    	angle $vangle
	location $fvup
	sky <0, 1, 0>
	look_at $sun_pos
	rotate -20 * x
	rotate 20 * y
    }

    light_source {
    	$fvup
	color White
    }
    
EOD

    for (my $i = 0; $i <= 9; $i++) {
    	my $pname = $planets[$i];
	if ($pname ne 'Moon') {
	    my $ppos = outpos($planet_vec{$pname});
	    my $psize = $planet_radius{$pname} * (($i > 0) ? $inflate_planets : 1);
	    my $orbital_radius = $planet_vec{$pname}->length;
	    
	    for (my $j = 0; $j <= 9; $j++) {
	    	if (($j != $i) && ($planets[$j] ne 'Moon')) {
		    my $oname = $planets[$j];
		    my $inorm = $planet_vec{$pname}->norm;
		    my $onorm = $planet_vec{$oname}->norm;
		    my $ixmag = ($inorm x $onorm)->length;
		    my $iang = asin($ixmag);
		    printf("%12s  %12s  %12.2f\n", $pname, $oname, rad2deg($iang));
		}
	    }
	    print("\n");
	    
	    
	    print OF << "EOD";
    //  $pname
    cylinder {
	<0, 0, 0>,     // Center of one end
	$ppos          // Center of other end
	0.150          // Radius
	texture {
	    pigment {
	    	color $planet_col{$pname}
	    }
	}
      }
            
      torus {
	  $orbital_radius, 0.1
	  rotate -90 * x
rotate 32 * y	    // HACK ATTACK
rotate 2.77 * z     // HACK ATTACK
	  pigment { color $planet_col{$pname} }
      }

      
      
      sphere {
    	$ppos, $psize
	texture {
	    pigment {
	    	color $planet_col{$pname}
	    }
	}
    }
    
EOD
 	    }
	}
	close(OF);
	
    }
    
    #	Render the scene for a given instant
    
    sub renderscene {
    	my ($fname) = @_;
	
	my $ofn = $fname;
	$ofn =~ s/pov$/png/;
	my $flabel = sprintf("%.2f  %4d-%02d-%02d %02d:%02d", $jd, $yy, $mo, $dd, $hh, $mm);
	my $ly = $fheight - 16;
	my $c = "povray $fname $pvopts +fp +o- | " .
		"ppmlabel -x 10 -y $ly -size 10 -text '$flabel' | " .
		"pnmtopng >$ofn\n";
	command($c);
    }
    
    #	Edit a vector to POV-Ray format

    sub outpos {
    	my ($p) = @_;
	
	return "< " . $p->x() . ", " . $p->y() . ", " . $p->z(). " >";
    }

    #	Print a command and then run it
    
    sub command {
    	my ($c) = @_;
	
	print($c);
	system($c);
    }
